'''
References
arxiv: quant-ph/0503237
Physical Review A 49, 1567 (1994)

Conversion between vectors of deffrent ordered operators
R = (q_1, p_1, ..., q_n, p_n)
S = (q_1, ..., q_n, p_1, ..., p_n)
T = (a_1, ..., a^*_1, ..., a^*n)
'''

import numpy as np

def StoTmat(cov):
    N2 = cov.shape[0]
    N = np.int(N2/2)
    Omega = np.zeros([N2, N2]) + 0j
    for i in range(N):
        Omega[i, i] = 1 / np.sqrt(2)
        Omega[i, i + N] = 1j / np.sqrt(2)
        Omega[i + N, i] = 1 / np.sqrt(2)
        Omega[i + N, i + N] = -1j / np.sqrt(2)

    cov_T = np.dot(Omega, np.dot(cov, np.transpose(np.conj(Omega))))
    return cov_T

def RtoSmat(cov):
    N2 = cov.shape[0]
    N = np.int64(N2/2)
    P = np.zeros([N2, N2]) + 0j
    for i in range(N):
        P[i, i * 2] = 1
    for i in range(N, N2):
        P[i, (i-N) * 2 + 1] = 1
    cov_T = np.dot(P, np.dot(cov, np.transpose(P)))
    return cov_T

def RtoTmat(cov):
    cov_T = RtoSmat(cov)
    return StoTmat(cov_T)

def StoTvec(vec):
    N2 = len(vec)
    N = np.int(N2/2)
    T = np.zeros(N2) + 0j
    T[:N] = (vec[:N] + 1j * vec[N:N2]) / np.sqrt(2)
    T[N:N2] = (vec[:N] - 1j * vec[N:N2]) / np.sqrt(2)
    return T

def RtoSvec(vec):
    N2 = len(vec)
    N = np.int(N2/2)
    S = np.zeros(N2) + 0j
    S[:N] = vec[0:N2:2]
    S[N:N2] = vec[1:N2:2]
    return S

def RtoTvec(vec):
    S = RtoSvec(vec)
    return StoTvec(S)

def RtoQmat(cov):
    N = np.int(cov.shape[0] / 2)
    V = RtoSmat(cov)
    Vxx = V[:N, :N]
    Vxp = V[:N, N:]
    Vpp = V[N:, N:]
    A = Vxx - 1j * (Vxp - Vxp.T) + Vpp
    B = Vxx + 1j * (Vxp - Vxp.T) + Vpp
    C = Vxx - 1j * (Vxp + Vxp.T) - Vpp
    sigma_Q = np.block([[A, C], [np.conj(C), B]]) * 0.5 + np.eye(2 * N) * 0.5
    return sigma_Q



import numpy as np
from scipy.special import factorial as fact
from scipy.linalg import expm
from sympy.physics.quantum import Operator
from sympy.parsing.sympy_parser import parse_expr
from sympy import Matrix, matrix2numpy, MatrixSymbol
from sympy import I as symI

def _downMat(dim, order):    
    if order == 0:
        A = np.eye(dim)
        return A
    else:
        A = np.zeros([dim, dim])
        for i in np.arange(order, dim):
            A[i, i - order] = np.prod(np.sqrt(np.arange(i, i - order, -1)))
        return A

def _upMat(dim, order):        
    if order == 0:
        A = np.eye(dim)
        return A
    else:
        A = np.zeros([dim, dim])
        for i in np.arange(0, dim - order):
            A[i, i + order] = np.prod(np.sqrt(np.arange(i + 1, i + 1 + order)))
        return A

def _downMatLeft(dim, order):    
    if order == 0:
        A = np.eye(dim)
        return A
    else:
        A = np.zeros([dim, dim])
        for i in np.arange(order, dim):
            A[i - order, i] = np.prod(np.sqrt(np.arange(i, i - order, -1)))
        return A

def _upMatLeft(dim, order):        
    if order == 0:
        A = np.eye(dim)
        return A
    else:
        A = np.zeros([dim, dim])
        for i in np.arange(0, dim - order):
            A[i - order, i] = np.prod(np.sqrt(np.arange(i + 1, i + 1 + order)))
        return A

def _nMat(dim, order):
    if order == 0:
        A = np.eye(dim)
        return A
    else:
        A = np.diag(np.arange(dim) ** order)
        return A

def exp_annihiration(fockState, alpha, order = 1, cutoff = 10):
    row = fockState.shape[0]
    mat = _downMat(fockState.shape[-1], order)
    mat_ = np.empty(mat.shape, dtype=np.complex)
    mat_ = expm(alpha * mat)
    res = np.dot(fockState, mat_)
    return res

def exp_creation(fockState, alpha, order = 1, cutoff = 10):
    row = fockState.shape[0]
    mat = _upMat(fockState.shape[-1], order)
    mat_ = np.empty(mat.shape, dtype=np.complex)
    mat_ = expm(alpha * mat)
    res = np.dot(fockState, mat_)
    return res

def exp_photonNum(fockState, alpha, order = 1, cutoff = 10):
    row = fockState.shape[0]
    mat = _nMat(fockState.shape[-1], order)
    mat_ = np.empty(mat.shape, dtype=np.complex)
    mat_ = expm(alpha * mat)
    res = np.dot(fockState, mat_)
    return res

def _mat_for_mode2(mat):
    l = mat.shape[0]
    mat_ = np.zeros(np.array(mat.shape)**2)
    for i in range(mat.shape[0]):
        mat_[i*l:i*l+l, i*l:i*l+l] = mat
    return mat_

def _mat_for_mode1(mat):
    l = mat.shape[0]
    mat_ = np.zeros(np.array(mat.shape)**2)
    for i in range(mat.shape[0]):
        for j in range(mat.shape[0]):
            mat_[i*l:i*l+l, j*l:j*l+l] = np.eye(l) * mat[i, j]
    return mat_

def exp_BS(fockState, alpha, cutoff):
    state = np.zeros(fockState.shape) + 0j
    down = _downMat(cutoff + 1, 1)
    up = _upMat(cutoff + 1, 1)
    mat1_ = np.dot(_mat_for_mode1(up), _mat_for_mode2(down))
    mat2_ = np.dot(_mat_for_mode1(down), _mat_for_mode2(up))
    mat_ = mat1_ - mat2_
    emat_ = expm(alpha * mat_)
    res = np.dot(fockState, emat_)
    return res

def exp_AAaa(fockState, alpha, cutoff):
    mat = _downMat(fockState.shape[-1], 2)
    #mat = np.dot(_upMat(fockState.shape[-1], 2), mat)
    mat = np.dot(mat, _upMat(fockState.shape[-1], 2))
    mat_ = np.empty(mat.shape, dtype=np.complex)
    mat_ = expm(alpha * mat)
    res = np.dot(fockState, mat_)
    return res

def exp_aa_minus_AA(fockState, alpha, cutoff):
    mat_a = _downMat(fockState.shape[-1], 2) # aa
    mat_A = _upMat(fockState.shape[-1], 2) # AA
    mat_ = np.empty(mat_a.shape, dtype=np.complex)
    mat_ = expm(np.conj(alpha) * mat_a - alpha * mat_A)
    res = np.dot(fockState, mat_)
    return res

def expand_xp_to_aA(expr_string, dim, evaluate = True, hbar = 1):
    x = Operator('x')
    p = Operator('p')
    expr = parse_expr(expr_string, local_dict = {'x':x, 'p':p})
    a_ = Operator('a')
    A_ = Operator('A')
    expr_ = expr.subs([(x, (a_ + A_) * sym.sqrt(hbar / 2)), (p, (a_ - A_) * sym.sqrt(hbar / 2) / 1j)])
    expr_ = expr_.expand()
    aA = str(expr_)
    if evaluate == False:
        return aA

    a = sym.MatrixSymbol('a', dim, dim)
    A = sym.MatrixSymbol('A', dim, dim)
    down = Matrix(downMatLeft(dim, 1))
    up = Matrix(upMatLeft(dim, 1))
    expr__ = parse_expr(aA, local_dict = {'a':a, 'A':A, 'I':sym.I})
    expr__ = expr__.subs([(a, down), (A, up)])
    res = matrix2numpy(expr__, dtype = np.complex)
    return res

def exp_str_aA(expr_string, alpha, N = 5):
    a = sym.Symbol('a', commutative = False)
    A = sym.Symbol('A', commutative = False)
    E = sym.Symbol('E', commutative = False)
    expr = parse_expr(expr_string, local_dict = {'a':a, 'A':A})
    exp_expr = E
    for i in range(1, N + 1):
        exp_expr += alpha**i * (expr)**i / np.math.factorial(i)
    exp_aA = exp_expr.expand()
    return str(exp_aA)
    
def str_to_aA_mat(expr_string, dim, hbar = 1):
    a = sym.MatrixSymbol('a', dim, dim)
    A = sym.MatrixSymbol('A', dim, dim)
    down = Matrix(_downMatLeft(dim, 1))
    up = Matrix(_upMatLeft(dim, 1))
    expr_ = parse_expr(expr_string, local_dict = {'a':a, 'A':A, 'I':sym.I, 'E':sym.eye(dim)})
    expr_ = expr_.subs([(a, down), (A, up)])
    res = matrix2numpy(expr_, dtype = np.complex)
    return res



"""
`gate` module implements quantum gate operations.
This module is internally used.
"""

import numpy as np
from .baseFunc import *
from .gateOps import *

class Sgate():
    """
    Squeezing gate.
    """
    def __init__(self, obj, idx, r):
        self.obj = obj
        self.N = self.obj.N
        self.idx = idx
        self.r = r

    def run(self, state):
        self.r = _paramCheck(self.r)
        return Xsqueeze(state, self.N, self.idx, self.r)

class PSgate():
    """
    Squeezing gate for p direction.
    """
    def __init__(self, obj, idx, r):
        self.obj = obj
        self.N = self.obj.N
        self.idx = idx
        self.r = r

    def run(self, state):
        self.r = _paramCheck(self.r)
        return Xsqueeze(state, self.N, self.idx, -self.r)

class Rgate():
    """
    Rotation gate.
    """
    def __init__(self, obj, idx, theta):
        self.obj = obj
        self.N = self.obj.N
        self.idx = idx
        self.theta = theta

    def run(self, state):
        self.theta = _paramCheck(self.theta)
        return Rotation(state, self.N, self.idx, self.theta)

class BSgate():
    """
    Beamsplitter gate.
    """
    def __init__(self, obj, idx1, idx2, theta):
        self.obj = obj
        self.N = self.obj.N
        self.idx1 = idx1
        self.idx2 = idx2
        self.theta = theta

    def run(self, state):
        self.theta = _paramCheck(self.theta)
        return Beamsplitter(state, self.N, self.idx1, self.idx2, self.theta)

class Dgate():
    """
    Displacement gate.
    """
    def __init__(self, obj, idx, alpha):
        self.obj = obj
        self.N = self.obj.N
        self.idx = idx
        self.alpha = alpha

    def run(self, state):
        self.alpha = _paramCheck(self.alpha)
        return Displace(state, self.idx, self.alpha)

class Xgate():
    """
    x axis shift in phase space.
    """
    def __init__(self, obj, idx, dx):
        self.obj = obj
        self.N = self.obj.N
        self.idx = idx
        self.dx = dx

    def run(self, state):
        self.dx = _paramCheck(self.dx)
        return Xtrans(state, self.idx, self.dx)

class Zgate():
    """
    p axis shift in phase space.
    """
    def __init__(self, obj, idx, dp):
        self.obj = obj
        self.N = self.obj.N
        self.idx = idx
        self.dp = dp

    def run(self, state):
        self.dp = _paramCheck(self.dp)
        return Ztrans(state, self.idx, self.dp)

class TMSgate():
    """
    Two mode squeezing gate.
    """
    def __init__(self, obj, idx1, idx2, r):
        self.obj = obj
        self.N = self.obj.N
        self.idx1 = idx1
        self.idx2 = idx2
        self.r = r

    def run(self, state):
        self.r = _paramCheck(self.r)
        return twoModeSqueezing(state, self.N, self.idx1, self.idx2, self.r)

class MeasX():
    """
    Homodyne measurement for x quadrature.
    """
    def __init__(self, obj, idx):
        self.obj = obj
        self.N = self.obj.N
        self.idx = idx

    def run(self, state):
        res, state_ = HomodyneX(state, self.idx)
        self.obj.creg[self.idx][0] = res
        return state_

class MeasP():
    """
    Homodyne measurement for p quadrature.
    """
    def __init__(self, obj, idx):
        self.obj = obj
        self.N = self.obj.N
        self.idx = idx

    def run(self, state):
        res, state_ = HomodyneP(state, self.idx)
        self.obj.creg[self.idx][1] = res
        return state_

def _paramCheck(param):
    if isinstance(param, CregReader):
        return param.read()
    else:
         return param

class CregReader():
    """
    Class for reading classical register.
    """
    def __init__(self, reg, idx, var, scale = 1):
        self.reg = reg
        self.idx = idx
        self.var = var
        self.scale = scale

    def read(self):
        if self.var == "x":
            v = 0
        elif self.var == "p":
            v = 1
        else:
            raise ValueError('Creg keeps measurement results of "x" or "p".')
        return self.reg[self.idx][v] * self.scale


"""This module is EXPERIMENTAL FEATURES. Provided functions may be removed or destructive changed.
This module provides functions that handles Circuits."""
from .circuit_to_unitary import circuit_to_unitary
from .flatten import flatten




"""
`gate` module implements quantum gate operations.
This module is internally used.
"""

import numpy as np
from .bosonicLadder import *
from .gateOps import *

class GATE():
    """Quantum gate class."""
    def __init__(self, obj):
        self.obj = obj

class Dgate(GATE):
    """
    Displacement gate.
    """
    def __init__(self, obj, mode, alpha):
        self.obj = obj
        self.cutoff = self.obj.cutoff
        self.N = self.obj.N
        self.mode = mode
        self.alpha = alpha
        super().__init__(obj)

    def run(self, state):
        self.alpha = _paramCheck(self.alpha)
        return Displacement(state, self.mode, self.alpha, self.N, self.cutoff)

class BSgate(GATE):
    """
    Beamsplitter gate.
    """    
    def __init__(self, obj, mode1, mode2, theta = np.pi / 4):
        self.obj = obj
        self.cutoff = self.obj.cutoff
        self.N = self.obj.N
        self.mode1 = mode1
        self.mode2 = mode2
        self.theta = theta
        super().__init__(obj)

    def run(self, state):
        self.theta = _paramCheck(self.theta)
        return Beamsplitter(state, self.mode1, self.mode2, self.theta, self.N, self.cutoff)

class Sgate(GATE):
    """
    Squeezing gate.
    """
    def __init__(self, obj, mode, r, phi = 0):
        self.obj = obj
        self.cutoff = self.obj.cutoff
        self.N = self.obj.N
        self.mode = mode
        self.r = r
        self.phi = phi
        super().__init__(obj)

    def run(self, state):
        self.r = _paramCheck(self.r)
        self.phi = _paramCheck(self.phi)
        return Squeeze(state, self.mode, self.r, self.phi, self.N, self.cutoff)

class Kgate(GATE):
    """
    Kerr gate.
    """    
    def __init__(self, obj, mode, chi):
        self.obj = obj
        self.cutoff = self.obj.cutoff
        self.N = self.obj.N
        self.mode = mode
        self.chi = chi
        super().__init__(obj)

    def run(self, state):
        self.chi = _paramCheck(self.chi)
        return KerrEffect(state, self.mode, self.chi, self.N, self.cutoff)

class polyH(GATE):
    """
    Time evolution for one qumode Hamiltonian.
    """    
    def __init__(self, obj, mode, gamma, expr):
        self.obj = obj
        self.cutoff = self.obj.cutoff
        self.N = self.obj.N
        self.mode = mode
        self.gamma = gamma
        self.expr = expr
        if not isinstance(expr, str):
            raise ValueError("Polynomial expression must be string.")
        super().__init__(obj)

    def run(self, state):
        self.gamma = _paramCheck(self.gamma)
        return HamiltonianEvo(state, self.mode, self.expr, self.gamma, self.N, self.cutoff)

class MeasF(GATE):
    """
    Photon number measurement gate.
    """
    def __init__(self, obj, mode, post_select = None):
        self.obj = obj
        self.cutoff = self.obj.cutoff
        self.N = self.obj.N
        self.mode = mode
        self.post_select = post_select
        super().__init__(obj)

    def run(self, state):
        res, state = photonMeasurement(state, self.mode, self.post_select)
        self.obj.creg[self.mode][2] = res
        return state

class MeasX(GATE):
    """
    Homodyne measurement gate.
    """
    def __init__(self, obj, mode, post_select = None):
        self.obj = obj
        self.cutoff = self.obj.cutoff
        self.N = self.obj.N
        self.mode = mode
        self.post_select = post_select
        super().__init__(obj)

    def run(self, state):
        res, state = homodyneMeasurement(state, self.mode, 0, self.post_select)
        self.obj.creg[self.mode][0] = res
        return state
        
class MeasP(GATE):
    """
    Homodyne measurement gate.
    """
    def __init__(self, obj, mode, post_select = None):
        self.obj = obj
        self.cutoff = self.obj.cutoff
        self.N = self.obj.N
        self.mode = mode
        self.post_select = post_select
        super().__init__(obj)

    def run(self, state):
        res, state = homodyneMeasurement(state, self.mode, np.pi / 2, self.post_select)
        self.obj.creg[self.mode][1] = res
        return state

def _paramCheck(param):
    if isinstance(param, CregReader):
        return param.read()
    else:
         return param

class CregReader():
    """
    Class for reading classical register.
    """
    def __init__(self, reg, idx, var, scale = 1):
        self.reg = reg
        self.idx = idx
        self.var = var
        self.scale = scale

    def read(self):
        if self.var == "x":
            v = 0
        elif self.var == "p":
            v = 1
        elif self.var == "n":
            v = 2
        else:
            raise ValueError('Creg keeps measurement results of "x" or "p" or "n".')
        return self.reg[self.idx][v] * self.scale


from .backendbase import Backend
from ..circuit import Circuit
from ..gate import *

import numpy as np


class Quimb(Backend):
    
    def _preprocess_run(self, gates, n_qubits, args, kwargs):
        import quimb.tensor as qtn
        
        circ = qtn.Circuit(n_qubits)
        regs = list(range(n_qubits))
        
        if "shots" in kwargs:
            n_shots = kwargs["shots"]
        else:
            n_shots = 1
        if "amplitude" in kwargs:
            amp = kwargs["amplitude"]
        else:
            amp = None
        if "hamiltonian" in kwargs:
            hami = kwargs["hamiltonian"]
        else:
            hami = None
        return gates, (circ, regs, n_qubits, n_shots, amp, hami)

    def _postprocess_run(self, ctx):
        import quimb as qu
        import collections
        import re
        
        if ctx[5]:
            hami = ctx[5].simplify()
            c = 0

            #because term not included into the array
            if len(re.split('[+,-]', str(hami))) == 1:
                kron_hami = qu.pauli(hami[0][0].op)
                qu_list = [hami[0][0][0]]
                for i in range(1, len(hami[0])):
                    kron_hami = qu.kron(kron_hami, qu.pauli(hami[0][i].op))
                    qu_list.append(hami[0][i][0])
                
                c += hami[1] * ctx[0].local_expectation(kron_hami, tuple(qu_list)).real
            
            else:
                for item in hami[0]:
                    if item[0]:
                        op = item[0][0].op
                        qubit = item[0][0][0]

                    else:
                        op = "I"
                        qubit = 0
                    kron_hami = qu.pauli(op)
                    qu_list = [qubit]
                
                    for i in range(1, len(item[0])):
                        if item[0]:
                            op = item[0][i].op
                            qubit = item[0][i][0]
                        else:
                            op = "I"
                            qubit = 0
                        kron_hami = qu.kron(kron_hami, qu.pauli(op))
                        qu_list.append(qubit)

                    c += item[1] * ctx[0].local_expectation(kron_hami, tuple(qu_list)).real
                
        #amplitude
        elif ctx[4]:
            c = ctx[0].amplitude(ctx[4])
            
        #sampling
        else:
            c = collections.Counter(ctx[0].sample(ctx[3]))

        return c

    def _one_qubit_gate_noargs(self, gate, ctx):
        for idx in gate.target_iter(ctx[2]):
            ctx[0].apply_gate(gate.lowername, ctx[1][idx])
        return ctx

    def _one_qubit_gate_args_theta(self, gate, ctx):
        for idx in gate.target_iter(ctx[2]):
            ctx[0].apply_gate(gate.lowername, gate.theta, ctx[1][idx])
        return ctx
    
    def _two_qubit_gate_noargs(self, gate, ctx):
        for control, target in gate.control_target_iter(ctx[2]):
            ctx[0].apply_gate(gate.lowername, ctx[1][control], ctx[1][target])
        return ctx

    def _two_qubit_gate_args_theta(self, gate, ctx):
        if gate.lowername == 'rxx':
            for control, target in gate.control_target_iter(ctx[2]):
                ctx[0].apply_gate('H', ctx[1][control])
                ctx[0].apply_gate('H', ctx[1][target])
                ctx[0].apply_gate('RZZ', gate.theta, ctx[1][control], ctx[1][target])
                ctx[0].apply_gate('H', ctx[1][control])
                ctx[0].apply_gate('H', ctx[1][target])
        elif gate.lowername == 'ryy':
            for control, target in gate.control_target_iter(ctx[2]):
                ctx[0].apply_gate('RX', -np.pi/4, ctx[1][control])
                ctx[0].apply_gate('RX', -np.pi/4, ctx[1][target])
                ctx[0].apply_gate('RZZ', gate.theta, ctx[1][control], ctx[1][target])
                ctx[0].apply_gate('RX', np.pi/4, ctx[1][control])
                ctx[0].apply_gate('RX', np.pi/4, ctx[1][target])
        else:
            for control, target in gate.control_target_iter(ctx[2]):
                ctx[0].apply_gate(gate.lowername, gate.theta, ctx[1][control], ctx[1][target])
        return ctx

    def _three_qubit_gate_noargs(self, gate, ctx):
        return ctx

    def gate_measure(self, gate, ctx):
        return ctx

    gate_x = gate_y = gate_z = gate_h = gate_t = gate_s = _one_qubit_gate_noargs
    gate_rx = gate_ry = gate_rz = gate_phase = _one_qubit_gate_args_theta
    gate_cx = gate_cy = gate_cz = _two_qubit_gate_noargs
    gate_rxx = gate_ryy = gate_rzz = _two_qubit_gate_args_theta
    gate_ccx = gate_cswap = _three_qubit_gate_noargs
    gate_reset = _one_qubit_gate_noargs


import numpy as np
import matplotlib.pyplot as plt
from .Fockbase.gates import *
from .Fockbase.gateOps import homodyneFock
from .Fockbase.states import *
from .Fockbase.WignerFunc import *

STATE_SET = {
    "vacuum": vacuumState,
    "coherent": coherentState,
    "cat": catState,
    "n_photon": photonNumberState
}

GATE_SET = {
    "D": Dgate,
    "BS": BSgate,
    "S": Sgate,    
    "Kerr": Kgate,
    "MeasF": MeasF,
    "MeasX": MeasX,
    "MeasP": MeasP,
    "polyH": polyH,
}

class Fock():
    """
    Class for continuous variable quantum compting in Fock basis.
    """
    def __init__(self, N, cutoff = 10):
        self.N = N
        self.cutoff = cutoff
        self.initState = np.zeros([N, self.cutoff + 1]) + 0j
        self.initState[:, 0] = 1
        self.state = None
        self.initStateFlag = False
        self.ops = []
        self.creg = [[None, None, None] for i in range(self.N)] # [x, p, n]

    def  __getattr__(self, name):
        if name in STATE_SET:
            if self.initStateFlag:
                raise ValueError("State must be set before gate operation.")
            self.ops.append(STATE_SET[name])
            return self._setGateParam

        elif name in GATE_SET:
            self.ops.append(GATE_SET[name])
            self.initStateFlag = True
            return self._setGateParam

        else:
            raise AttributeError('The method does not exist')

    def _setGateParam(self, *args, **kwargs):
        self.ops[-1] = self.ops[-1](self, *args, **kwargs)
        return self

    def Creg(self, idx, var, scale = 1):
        """
        Access to classical register.
        """
        return CregReader(self.creg, idx, var, scale)

    def run(self):
        """
        Run the circuit.
        """
        for op in self.ops:
            if isinstance(op, STATE):
                self.initState = op.run(self.initState)
            elif isinstance(op, GATE):
                if self.state is None:
                    self.state = self._multiTensordot()
                self.state = op.run(self.state)
                sum_of_prob = np.sum(np.abs(self.state)**2)
                if np.abs(1 - sum_of_prob) < 0.3:
                    self.state /= np.sqrt(sum_of_prob)
                # print(op, np.sum(np.abs(self.state)**2))
        return self

    def _multiTensordot(self):
        self.state = self.initState[0, :]
        for i in range(self.N - 1):
            self.state = np.tensordot(self.state, self.initState[i+1, :], axes = 0)
        return self.state

    def Wigner(self, mode, method = 'clenshaw', plot = 'y', xrange = 5.0, prange = 5.0):
        """
        Calculate the Wigner function of a selected mode.
        
        Args:
            mode (int): Selecting a optical mode.
            method: "clenshaw" (default) or "moyal".
            plot: If 'y', the plot of wigner function is output using matplotlib.\
                 If 'n', only the meshed values are returned.
            x(p)range: The range in phase space for calculateing Wigner function.
        """
        if self.state is None:
            self.state = self._multiTensordot()
            self.initState == None
        x = np.arange(-xrange, xrange, xrange / 50)
        p = np.arange(-prange, prange, prange / 50)
        m = len(x)
        xx, pp = np.meshgrid(x, -p)
        W = FockWigner(xx, pp, self.state, mode, method)
        if plot == 'y':
            h = plt.contourf(x, p, W)
            plt.show()
        return (x, p, W)

    def photonSampling(self, mode, ite = 1):
        """
        Simulate the result of photon number resolving measurement for a selected mode.
        
        Args:
            mode (int): Selecting a optical mode.
            ite: The number of sampling.        
        """
        if self.state is None:
            self.state = self._multiTensordot()
            self.initState == None
        reducedDensity = reduceState(self.state, mode)
        probs = np.real(np.diag(reducedDensity))
        probs = probs / np.sum(probs)
        return np.random.choice(probs.shape[0], ite, p = probs)

    def homodyneSampling(self, mode, theta, ite = 1):
        if self.state is None:
            self.state = self._multiTensordot()
            self.initState == None
        res, psi = homodyneFock(self.state, mode, theta, ite = ite)
        return res



from .. import Circuit
from .. import gate as g


def flatten(c: Circuit) -> Circuit:
    """expands slice and multiple targets into single target"""
    n_qubits = c.n_qubits
    ops = []
    for op in c.ops:
        if isinstance(op, (g.OneQubitGate, g.Reset)):
            ops += [op.create(t, op.params, None) for t in op.target_iter(n_qubits)]
        elif isinstance(op, g.TwoQubitGate):
            ops += [
                op.create(t, op.params, None)
                for t in op.control_target_iter(n_qubits)
            ]
        elif isinstance(op, g.Measurement):
            if op.key is None:
                ops += [
                    op.create(t, op.params, None) for t in op.target_iter(n_qubits)
                ]
            else:
                options = {'key': op.key}
                if op.duplicated is not None:
                    options['duplicated'] = op.duplicated
                ops += [
                    op.create(tuple(t for t in op.target_iter(n_qubits)),
                              op.params, options)
                ]
        else:
            raise ValueError(f"Cannot process operation {op.lowername}.")
    return Circuit(n_qubits, ops)



"""This module is experiment features. We may delete or change provided functions.

The module provides builtin macros."""

import math
from typing import Any, Sequence, Optional

from blueqat import Circuit
from blueqat.decorators import circuitmacro
from blueqat.utils import calc_u_params, gen_gray_controls, sqrt_2x2_matrix

from blueqat.gate import UGate


@circuitmacro
def draw(c: Circuit, **kwargs) -> Any:
    """Draw the circuit. This function requires Qiskit."""
    return c.run(backend='ibmq', returns="draw", **kwargs)


@circuitmacro
def margolus(c: Circuit, c0: int, c1: int) -> Circuit:
    """Simplified Toffoli gate implementation by Margolus.

    This gate is also as know as relative Toffoli gate.
    It's almost as same as Toffoli gate, but only relative phases are different.
    Refer https://arxiv.org/abs/quant-ph/0312225v1 for details.
    (Remarks: It's also described in Nielsen & Chuang, Exercise 4.26.)"""
    c.ry(math.pi * 0.25)[t].cx[c1, t].ry(math.pi * 0.25)[t].cx[c0, t]
    c.ry(math.pi * -0.25)[t].cx[c1, t].ry(math.pi * -0.25)[t]
    return c


@circuitmacro
def c3z(c: Circuit, c0: int, c1: int, c2: int, t: int) -> Circuit:
    """A macro of 3-controlled Z gate"""
    return c.mcz_gray([c0, c1, c2], t)


@circuitmacro
def c4z(c: Circuit, c0: int, c1: int, c2: int, c3: int, t: int) -> Circuit:
    """A macro of 4-controlled Z gate"""
    return c.mcz_gray([c0, c1, c2, c3], t)


@circuitmacro
def c3x(c: Circuit, c0: int, c1: int, c2: int, t: int) -> Circuit:
    """A macro of 3-controlled X gate"""
    return c.h[t].c3z(c0, c1, c2, t).h[t]


@circuitmacro
def c4x(c: Circuit, c0: int, c1: int, c2: int, c3: int, t: int) -> Circuit:
    """A macro of 4-controlled X gate"""
    return c.h[t].c4z(c0, c1, c2, c3, t).h[t]


@circuitmacro
def mcx_gray(c: Circuit, ctrl: Sequence[int], target: int) -> Circuit:
    """A macro of multi controlled X gate."""
    return c.h[target].mcz_gray(ctrl, target).h[target]


@circuitmacro
def mcx_with_ancilla(c: Circuit,
                     ctrl: Sequence[int],
                     target: int,
                     ancilla: int) -> Circuit:
    """A macro of multi controlled X gate with ancilla.

    Refer https://arxiv.org/pdf/quant-ph/9503016.pdf Lem. 7.3, Cor. 7.4."""
    n_ctrl = len(ctrl)
    if n_ctrl == 0:
        return c.x[target]
    if n_ctrl == 1:
        return c.cx[ctrl[0], target]
    if n_ctrl == 2:
        return c.ccx[ctrl[0], ctrl[1], target]
    if n_ctrl == 3:
        return c.c3x(ctrl[0], ctrl[1], ctrl[2], target)
    if n_ctrl == 4:
        return c.c4x(ctrl[0], ctrl[1], ctrl[2], ctrl[3], target)
    sep = (n_ctrl + 1) // 2 + 1
    c1 = ctrl[:sep]
    a1 = ctrl[-1]
    c2 = list(ctrl[sep:]) + [ancilla]
    a2 = ctrl[sep - 1]
    c.mcx_with_ancilla(c1, ancilla, a1)
    c.mcx_with_ancilla(c2, target, a2)
    c.mcx_with_ancilla(c1, ancilla, a1)
    c.mcx_with_ancilla(c2, target, a2)
    return c


@circuitmacro
def mcz_gray(c: Circuit, ctrl: Sequence[int], target: int) -> Circuit:
    """A macro of multi controlled Z gate."""
    n_ctrl = len(ctrl)
    if n_ctrl == 0:
        return c.z[target]
    angles = [math.pi / 2**(n_ctrl - 1), -math.pi / 2**(n_ctrl - 1)]
    for c0, c1, parity in gen_gray_controls(n_ctrl):
        if c0 >= 0:
            c.cx[ctrl[c0], ctrl[c1]]
        c.cr(angles[parity])[ctrl[c1], target]
    return c


@circuitmacro
def mcz_with_ancilla(c: Circuit,
                     ctrl: Sequence[int],
                     target: int,
                     ancilla: int) -> Circuit:
    """A macro of multi controlled X gate with ancilla."""
    return c.h[target].mcx_with_ancilla(ctrl, target, ancilla).h[target]


@circuitmacro
def mcrx_gray(c: Circuit, theta: float, ctrl: Sequence[int],
              target: int) -> Circuit:
    """A macro of multi controlled RX gate."""
    n_ctrl = len(ctrl)
    if n_ctrl == 0:
        return c.rx(theta)[target]
    angles = [theta / 2**(n_ctrl - 1), -theta / 2**(n_ctrl - 1)]
    for c0, c1, parity in gen_gray_controls(n_ctrl):
        if c0 >= 0:
            c.cx[ctrl[c0], ctrl[c1]]
        c.crx(angles[parity])[ctrl[c1], target]
    return c


@circuitmacro
def mcry_gray(c: Circuit, theta: float, ctrl: Sequence[int],
              target: int) -> Circuit:
    """A macro of multi controlled RY gate."""
    n_ctrl = len(ctrl)
    if n_ctrl == 0:
        return c.ry(theta)[target]
    angles = [theta / 2**(n_ctrl - 1), -theta / 2**(n_ctrl - 1)]
    for c0, c1, parity in gen_gray_controls(n_ctrl):
        if c0 >= 0:
            c.cx[ctrl[c0], ctrl[c1]]
        c.cry(angles[parity])[ctrl[c1], target]
    return c


@circuitmacro
def mcrz_gray(c: Circuit, theta: float, ctrl: Sequence[int],
              target: int) -> Circuit:
    """A macro of multi controlled RZ gate."""
    n_ctrl = len(ctrl)
    if n_ctrl == 0:
        return c.rz(theta)[target]
    angles = [theta / 2**(n_ctrl - 1), -theta / 2**(n_ctrl - 1)]
    for c0, c1, parity in gen_gray_controls(n_ctrl):
        if c0 >= 0:
            c.cx[ctrl[c0], ctrl[c1]]
        c.crz(angles[parity])[ctrl[c1], target]
    return c


@circuitmacro
def mcr_gray(c: Circuit, theta: float, ctrl: Sequence[int],
              target: int) -> Circuit:
    """A macro of multi controlled R gate."""
    n_ctrl = len(ctrl)
    if n_ctrl == 0:
        return c.r(theta)[target]
    angles = [theta / 2**(n_ctrl - 1), -theta / 2**(n_ctrl - 1)]
    for c0, c1, parity in gen_gray_controls(n_ctrl):
        if c0 >= 0:
            c.cx[ctrl[c0], ctrl[c1]]
        c.cr(angles[parity])[ctrl[c1], target]
    return c


@circuitmacro
def mcu_gray(c: Circuit, theta: float, phi: float, lam: float, gamma: float,
             ctrl: Sequence[int], target: int) -> Circuit:
    """A macro of multi controlled U gate."""
    n_ctrl = len(ctrl)
    if n_ctrl == 0:
        return c.u(theta, phi, lam, gamma)[target]
    if n_ctrl == 1:
        return c.cu(theta, phi, lam, gamma)[ctrl[0], target]
    mat = UGate.create(0, (theta, phi, lam, gamma)).matrix()
    for _ in range(n_ctrl - 1):
        mat = sqrt_2x2_matrix(mat)
    params = calc_u_params(mat)
    params_dagger = -params[0], -params[2], -params[1], -params[3]
    for c0, c1, parity in gen_gray_controls(n_ctrl):
        if c0 >= 0:
            c.cx[ctrl[c0], ctrl[c1]]
        if parity:
            c.cu(*params_dagger)[ctrl[c1], target]
        else:
            c.cu(*params)[ctrl[c1], target]
    return c



"""Defines JSON serializer and deserializer."""
import typing
from blueqat import Circuit

from ..gateset import create
from .flatten import flatten

from blueqat.gate import Measurement, Operation

SCHEMA_NAME = 'blueqat-circuit'
AVAILABLE_SCHEMA_VERSIONS = ["1", "2"]
LATEST_SCHEMA_VERSION = "2"

if typing.TYPE_CHECKING:
    from typing import Any, Dict, List, Union
    try:
        from typing import TypedDict
    except ImportError:
        CircuitJsonDictV2 = Dict[str, Any]
        CircuitJsonDictV1 = Dict[str, Any]
        OpJsonDictV2 = Dict[str, Any]
        OpJsonDictV1 = Dict[str, Any]
    else:
        class SchemaJsonDict(TypedDict):
            """Schema header for detect data type"""
            name: str
            version: str

        class OpJsonDictV1(TypedDict):
            """Data type of Operation"""
            name: str
            params: List[float]
            targets: List[int]

        class OpJsonDictV2(TypedDict):
            """Data type of Operation"""
            name: str
            params: List[float]
            options: Dict[str, Any]
            targets: List[int]

        class CircuitJsonDictV1(TypedDict):
            """Data type of Circuit"""
            schema: SchemaJsonDict
            n_qubits: int
            ops: List[OpJsonDictV1]

        class CircuitJsonDictV2(TypedDict):
            """Data type of Circuit"""
            schema: SchemaJsonDict
            n_qubits: int
            ops: List[OpJsonDictV2]

    CircuitJsonDict = Union[CircuitJsonDictV1, CircuitJsonDictV2]


def serialize(c: Circuit) -> 'CircuitJsonDictV2':
    """Serialize Circuit into JSON type dict.

    In this implementation, serialized circuit is flattened.
    However, it's not specifications of JSON schema.
    """
    def serialize_op(op: 'Operation') -> 'OpJsonDictV2':
        targets = op.targets
        if isinstance(targets, slice):
            raise TypeError('Not flatten circuit.')
        if isinstance(targets, int):
            targets = [targets]
        if isinstance(targets, tuple):
            targets = list(targets)
        options = {}
        if isinstance(op, Measurement):
            if op.key is not None:
                options['key'] = op.key
            if op.duplicated is not None:
                options['duplicated'] = op.duplicated
        return {
            'name': str(op.lowername),
            'params': [float(p) for p in op.params],
            'options': options,
            'targets': targets
        }

    c = flatten(c)
    return {
        'schema': {
            'name': SCHEMA_NAME,
            'version': LATEST_SCHEMA_VERSION
        },
        'n_qubits': c.n_qubits,
        'ops': [serialize_op(op) for op in c.ops]
    }


def deserialize(data: 'CircuitJsonDict') -> Circuit:
    """Deserialize JSON type dict into Circuit"""
    def make_op(opdata: 'Union[OpJsonDictV1, OpJsonDictV2]') -> 'Operation':
        return create(opdata['name'],
                      tuple(opdata['targets']),
                      tuple(float(p) for p in opdata['params']),
                      opdata.get('options'))
    schema = data.get('schema', {})
    if schema.get('name', '') != SCHEMA_NAME:
        raise ValueError('Invalid schema')
    if schema.get('version', '') not in AVAILABLE_SCHEMA_VERSIONS:
        raise ValueError('Unknown schema version')
    n_qubits = data['n_qubits']
    ops = data['ops']
    return Circuit(n_qubits, [
        make_op(opdata) for opdata in ops
    ])



import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("blueqat/_version.py", "r", encoding="utf-8") as f:
    exec(f.read())

with open("requirements.txt", "r", encoding="utf-8") as f:
    install_requires = list(map(str.strip, f))

setuptools.setup(
    name = "blueqat",
    version=__version__,
    author="The Blueqat Developers",
    author_email="info@blueqat.com",
    description="Quantum Computer Library for Everyone",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Blueqat/Blueqat",
    license="Apache 2",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 3 - Alpha",
    ]
)





"""
`WignerFunc` module implements calculation of Wigner function.
This module is internally used.
"""

import numpy as np
from scipy.special import factorial as fact

def FockWigner(xmat, pmat, fockState, mode, method = 'clenshaw', tol=1e-10):
    if fockState.ndim < mode + 1:
        raise  ValueError("The mode is not exist.")
    if fockState.ndim > 1:
        rho = reduceState(fockState, mode)
    else:
        rho = np.outer(np.conj(fockState), fockState)
    if method == 'moyal':
        W = _Wigner_Moyal(rho, xmat, pmat, tol)
    elif method == 'clenshaw':
        W = _Wigner_clenshaw(rho, xmat, pmat, tol)
    else:
        raise ValueError("method is invalid.")
    return W

def reduceState(fockState, mode):
    modeNum = fockState.ndim
    cutoff = fockState.shape[-1] - 1
    fockState = np.swapaxes(fockState, mode, -1)
    fockState = fockState.flatten()
    rho = np.outer(np.conj(fockState), fockState)
    for i in range(modeNum - 1):
        rho = partialTrace(rho, cutoff)
    return rho

def partialTrace(rho, cutoff):
    dim1 = np.int(cutoff + 1)
    dim2 = np.int(rho.shape[0] / dim1)
    rho_ = np.zeros([dim2, dim2]) + 0j
    for j in range(dim1):
        rho_ += rho[(j * dim2):(j * dim2 + dim2), (j * dim2):(j * dim2 + dim2)]
    return rho_

def _Wigner_Moyal(rho, xmat, pmat, tol):
    dim = rho.shape[0]
    [l, m] = np.indices([dim, dim])
    A = np.max(np.dstack([l, m]), axis=2)
    B = np.abs(l - m)
    C = np.min(np.dstack([l, m]), axis=2)
    R0 = xmat**2 + pmat**2
    xmat = xmat[:, :, np.newaxis, np.newaxis]
    pmat = pmat[:, :, np.newaxis, np.newaxis]
    R = xmat**2 + pmat**2
    X = xmat - np.sign(l-m) * 1j * pmat
    W = 2 * (-1)**C * np.sqrt(2**(B) * fact(C) / fact(A))
    W = W * np.exp(-R) * X**(B)
    S = _Sonin(C, B, 2 * R0)
    W = W * S
    W = rho * W
    W = np.sum(np.sum(W, axis = -1), axis = -1)
    if np.max(np.imag(W)) < tol:
        W = np.real(W)
    else:
        raise ValueError("Wigner plot has imaginary value.")
    return W

# Based on QuTiP
def _Wigner_clenshaw(rho, xmat, pmat, tol, hbar = 1):
    g = np.sqrt(2 / hbar)
    M = rho.shape[0]
    A2 = g * (xmat + 1.0j * pmat)    
    B = np.abs(A2)
    B *= B
    w0 = (2*rho[0, -1])*np.ones_like(A2)
    L = M-1
    rho = rho * (2*np.ones((M,M)) - np.diag(np.ones(M)))
    while L > 0:
        L -= 1
        w0 = _Wigner_laguerre(L, B, np.diag(rho, L)) + w0 * A2 * (L+1)**-0.5
    W = w0 * np.exp(-B * 0.5) * (g ** 2 * 0.5 / np.pi)
    # if np.max(np.imag(W)) < tol:
    #     W = np.real(W)
    # else:
    #     raise ValueError("Wigner plot has imaginary value.")
    W = np.real(W)
    return W

def _Wigner_laguerre(L, x, c):
    
    if len(c) == 1:
        y0 = c[0]
        y1 = 0
    elif len(c) == 2:
        y0 = c[0]
        y1 = c[1]
    else:
        k = len(c)
        y0 = c[-2]
        y1 = c[-1]
        for i in range(3, len(c) + 1):
            k -= 1
            y0,    y1 = c[-i] - y1 * (float((k - 1)*(L + k - 1))/((L+k)*k))**0.5, \
            y0 - y1 * ((L + 2*k -1) - x) * ((L+k)*k)**-0.5
            
    return y0 - y1 * ((L + 1) - x) * (L + 1)**-0.5

def _to_2d_ndarray(a):
    if isinstance(a,(np.ndarray)):
        return a
    else:
        return np.array([[a]])

# slow!
def _Sonin(n, alpha, x):
    n = _to_2d_ndarray(n)
    alpha = _to_2d_ndarray(alpha)
    x = _to_2d_ndarray(x)
    a = fact(n + alpha)
    k0 = np.arange(np.max(n) + 1)
    k0 = k0[:, np.newaxis, np.newaxis]
    k = k0 * np.ones([np.max(n) + 1, n.shape[0], n.shape[0]], dtype = np.int)
    mask = np.ones(k.shape, dtype = np.int)
    for i in range(k.shape[0]):
        ind = (np.ones(n.shape) * i) > n
        mask[i, ind] = 0
    k *= mask
    S = mask * (-1)**k * a / fact(n - k) / fact(k + alpha) / fact(k)
    X = x ** k0
    S = S[:, np.newaxis, np.newaxis, :, :] * X[:, :, :, np.newaxis, np.newaxis]
    return np.sum(S, axis = 0)